from datetime import datetime
from typing import Dict, Callable, List, Tuple
from decimal import Decimal

from moneywiz_api.database_accessor import DatabaseAccessor
from moneywiz_api.model.transaction import (
    Transaction,
    DepositTransaction,
    InvestmentExchangeTransaction,
    InvestmentBuyTransaction,
    InvestmentSellTransaction,
    ReconcileTransaction,
    RefundTransaction,
    TransferBudgetTransaction,
    TransferDepositTransaction,
    TransferWithdrawTransaction,
    WithdrawTransaction,
)
from moneywiz_api.managers.record_manager import RecordManager
from moneywiz_api.types import ID


class TransactionManager(RecordManager[Transaction]):
    def __init__(self):
        super().__init__()
        self.category_assignment: Dict[ID, List[Tuple[ID, Decimal]]] = {}
        self.refund_maps: Dict[ID, ID] = {}
        self.tags_map: Dict[ID, ID] = {}

    @property
    def ents(self) -> Dict[str, Callable]:
        return {
            "DepositTransaction": DepositTransaction,
            "InvestmentExchangeTransaction": InvestmentExchangeTransaction,
            "InvestmentBuyTransaction": InvestmentBuyTransaction,
            "InvestmentSellTransaction": InvestmentSellTransaction,
            "ReconcileTransaction": ReconcileTransaction,
            "RefundTransaction": RefundTransaction,
            "TransferBudgetTransaction": TransferBudgetTransaction,
            "TransferDepositTransaction": TransferDepositTransaction,
            "TransferWithdrawTransaction": TransferWithdrawTransaction,
            "WithdrawTransaction": WithdrawTransaction,
        }

    def load(self, db_accessor: DatabaseAccessor) -> None:
        super().load(db_accessor)
        self.category_assignment: Dict[ID, List[Tuple[ID, Decimal]]] = (
            db_accessor.get_category_assignment()
        )
        self.refund_maps: Dict[ID, ID] = db_accessor.get_refund_maps()
        self.tags_map: Dict[ID, ID] = db_accessor.get_tags_map()

    def category_for_transaction(
        self, transaction_id: ID
    ) -> List[Tuple[ID, Decimal]] | None:
        return self.category_assignment.get(transaction_id)

    def tags_for_transaction(self, transaction_id: ID) -> List[ID] | None:
        return self.tags_map.get(transaction_id)

    def original_transaction_for_refund_transaction(
        self, transaction_id: ID
    ) -> ID | None:
        return self.refund_maps.get(transaction_id)

    def get_all_for_account(
        self, account_id: ID, until: datetime = datetime.now()
    ) -> List[Transaction]:
        """
        Get all transactions for a given account
        :param account_id:
        :param until: inclusive
        :return:
        """
        return sorted(
            [
                x
                for _, x in self.records().items()
                if not isinstance(x, TransferBudgetTransaction)
                and x.account == account_id
                and x.datetime <= until
            ],
            key=lambda x: x.datetime,
        )

    def get_all(self, until: datetime = datetime.now()) -> List[Transaction]:
        return sorted(
            [
                x
                for _, x in self.records().items()
                if not isinstance(x, TransferBudgetTransaction) and x.datetime <= until
            ],
            key=lambda x: x.datetime,
        )



from typing import Dict, Callable, List
from decimal import Decimal

from moneywiz_api.model.investment_holding import InvestmentHolding
from moneywiz_api.managers.record_manager import RecordManager
from moneywiz_api.types import ID


class InvestmentHoldingManager(RecordManager[InvestmentHolding]):
    def __init__(self):
        super().__init__()

    @property
    def ents(self) -> Dict[str, Callable]:
        return {
            "InvestmentHolding": InvestmentHolding,
        }

    def get_holdings_for_account(self, account_id: ID) -> List[InvestmentHolding]:
        return [x for _, x in self.records().items() if x.account == account_id]

    def update_last_price(self, latest_price: Decimal):
        raise NotImplementedError()

    def update_price_table(self, latest_price: Decimal):
        raise NotImplementedError()



from typing import Dict, Callable, List

from moneywiz_api.model.account import Account
from moneywiz_api.managers.record_manager import RecordManager

from moneywiz_api.model.account import (
    BankChequeAccount,
    BankSavingAccount,
    CashAccount,
    CreditCardAccount,
    LoanAccount,
    InvestmentAccount,
    ForexAccount,
)
from moneywiz_api.types import ID, GID


class AccountManager(RecordManager[Account]):
    def __init__(self):
        super().__init__()

    @property
    def ents(self) -> Dict[str, Callable]:
        return {
            "BankChequeAccount": BankChequeAccount,
            "BankSavingAccount": BankSavingAccount,
            "CashAccount": CashAccount,
            "CreditCardAccount": CreditCardAccount,
            "LoanAccount": LoanAccount,
            "InvestmentAccount": InvestmentAccount,
            "ForexAccount": ForexAccount,
        }

    def records(self) -> Dict[ID, Account]:
        return dict(sorted(super().records().items(), key=lambda x: x[1].display_order))

    def get_accounts_for_user(self, user_id: ID) -> List[Account]:
        return sorted(
            [x for _, x in self.records().items() if x.user == user_id],
            key=lambda x: (x.group_id, x.display_order),
        )



from typing import Dict, Callable, List

from moneywiz_api.model.category import Category
from moneywiz_api.managers.record_manager import RecordManager
from moneywiz_api.types import ID, GID


class CategoryManager(RecordManager[Category]):
    def __init__(self):
        super().__init__()

    @property
    def ents(self) -> Dict[str, Callable]:
        return {
            "Category": Category,
        }

    def get_name_chain(self, category_id: ID) -> List[str]:
        ret: List[str] = []
        current = self.get(category_id)
        while current:
            ret.insert(0, current.name)
            if not current.parent_id:
                break
            else:
                current = self.get(current.parent_id)
        return ret

    def get_name_chain_by_gid(self, category_gid: GID) -> List[str]:
        current = self.get_by_gid(category_gid)
        return self.get_name_chain(current.id)

    def get_categories_for_user(self, user_id: ID) -> List[Category]:
        return sorted(
            [x for _, x in self.records().items() if x.user == user_id],
            key=lambda x: x.type,
        )



from datetime import datetime

from moneywiz_api.cli.cli import get_default_path

TEST_DB_PATH = get_default_path()

BALANCE_AS_OF_DATE = datetime(2023, 5, 19, 0, 0, 0)
CASH_BALANCES = [
    # (ACCOUNT_PK, BALANCE)
    (1001, -100.00),
    (1002, -202.33),
]

HOLDINGS_BALANCES = [
    # (ACCOUNT_PK, {
    #     HOLDINGS_PK: HOLDINGS_BALANCE)
    # })
    (
        2001,
        {
            3001: 15,
        },
    ),
]



from dataclasses import dataclass, asdict, field
from typing import Dict, Any
from datetime import datetime

from moneywiz_api.types import ID, ENT_ID
from moneywiz_api.model.raw_data_handler import RawDataHandler as RDH


@dataclass
class Record:
    _raw: Dict[str, Any] = field(repr=False)
    _ent: ENT_ID = field(repr=False)
    _created_at: datetime = field(repr=False)
    gid: str = field(repr=False)
    id: ID

    def __init__(self, row):
        self._raw = row
        self._ent = row["Z_ENT"]
        self._created_at = RDH.get_datetime(row, "ZOBJECTCREATIONDATE")
        self.gid = row["ZGID"]
        self.id = row["Z_PK"]

        # Fixes

        # Validate
        assert self._raw
        assert self._ent
        assert self._created_at
        assert self.gid
        assert self.id

    def ent(self) -> ENT_ID:
        return self._ent

    def filtered(self) -> Dict[str, Any]:
        """
        Utility function to return cleaned up entities.
        it will exclude fields like binary, Z9_

        :return:
        """
        return RDH.filter_row(self._raw)
        # copy = {k: v for k, v in self._raw.items()}
        # # del copy["ZGID"]
        # del copy["ZMANUALHISTORICALPRICESPERSHARE"]
        # del copy["ZIMPORTLINKIDARRAY2"]
        # del copy["ZIMPORTLINKIDARRAY"]
        # del copy["ZBANKLOGOPRIMARYCOLOR"]
        # return {
        #     k: v
        #     for k, v in copy.items()
        #     if (v is not None) and (not k.startswith("Z9_"))
        # }

    def as_dict(self) -> Dict[str, Any]:
        """
        Utility function to return dataclass instance as a dict.

        :return:
        """
        original = asdict(self)
        del original["_raw"]
        del original["_ent"]
        del original["_created_at"]
        return original



from dataclasses import dataclass

from moneywiz_api.model.record import Record
from moneywiz_api.types import ID


@dataclass
class Payee(Record):
    """
    ENT: 28
    """

    name: str
    user: ID

    def __init__(self, row):
        super().__init__(row)
        self.name = row["ZNAME5"]
        self.user = row["ZUSER7"]

        # Fixes

        # Validate
        assert self.name is not None, self.as_dict()
        assert self.user is not None, self.as_dict()



import sqlite3
from collections import defaultdict
from pathlib import Path
from typing import Dict, List, Any, Callable, Tuple
from decimal import Decimal

from moneywiz_api.model.record import Record
from moneywiz_api.model.raw_data_handler import RawDataHandler as RDH
from moneywiz_api.types import ENT_ID, ID, GID


class DatabaseAccessor:
    def __init__(self, db_path: Path):
        self._con = sqlite3.connect(db_path, uri=True)

        def dict_factory(cursor, row):
            record = {}
            for idx, col in enumerate(cursor.description):
                record[col[0]] = row[idx]
            return record

        self._con.row_factory = dict_factory

        self._ent_to_typename: Dict[ENT_ID, str] = self._load_primarykey()
        self._typename_to_ent: Dict[str, ENT_ID] = {
            v: k for k, v in self._ent_to_typename.items()
        }

    def _load_primarykey(self) -> Dict[int, str]:
        cur = self._con.cursor()
        res = cur.execute(
            """
        SELECT * FROM  "Z_PRIMARYKEY" ORDER BY "Z_ENT" LIMIT 1000 OFFSET 0;
        """
        )
        ent_to_typename: Dict[int, str] = {}
        for row in res.fetchall():
            ent_to_typename[row["Z_ENT"]] = row["Z_NAME"]
        return ent_to_typename

    def __repr__(self):
        return "\n".join(
            f"{key}: {value}" for key, value in self._ent_to_typename.items()
        )

    def typename_for(self, ent_id: ENT_ID) -> str:
        return self._ent_to_typename.get(ent_id)

    def ent_for(self, typename: str) -> ENT_ID:
        return self._typename_to_ent.get(typename)

    def query_objects(self, typenames: List[str]) -> List[Any]:
        cur = self._con.cursor()
        res = cur.execute(
            """
        SELECT * FROM ZSYNCOBJECT WHERE Z_ENT in (%s)
        """
            % (",".join("?" * len(typenames))),
            [self.ent_for(x) for x in typenames],
        )
        return res.fetchall()

    def get_record(self, pk_id: ID, constructor: Callable = Record):
        cur = self._con.cursor()
        res = cur.execute(
            """
        SELECT * FROM ZSYNCOBJECT WHERE Z_PK = ?
        
        """,
            [pk_id],
        )

        return constructor(res.fetchone())

    def get_record_by_gid(self, gid: GID, constructor: Callable = Record):
        cur = self._con.cursor()
        res = cur.execute(
            """
        SELECT * FROM ZSYNCOBJECT WHERE ZGID = ?
        
        """,
            [gid],
        )

        return constructor(res.fetchone())

    def get_category_assignment(self) -> Dict[ID, List[Tuple[ID, Decimal]]]:
        transaction_map: Dict[ID, List[Tuple[ID, Decimal]]] = defaultdict(list)
        cur = self._con.cursor()
        res = cur.execute(
            """
        SELECT ZCATEGORY, ZTRANSACTION, ZAMOUNT  FROM ZCATEGORYASSIGMENT WHERE ZTRANSACTION IS NOT NULL
        
        """
        )
        for row in res.fetchall():
            transaction_map[row["ZTRANSACTION"]].append(
                (row["ZCATEGORY"], RDH.get_decimal(row, "ZAMOUNT"))
            )
        return transaction_map

    def get_refund_maps(self) -> Dict[ID, ID]:
        refund_to_withdraw: Dict[ID, ID] = {}
        cur = self._con.cursor()
        res = cur.execute(
            """
        SELECT ZREFUNDTRANSACTION, ZWITHDRAWTRANSACTION  FROM ZWITHDRAWREFUNDTRANSACTIONLINK
        
        """
        )
        for row in res.fetchall():
            refund_to_withdraw[row["ZREFUNDTRANSACTION"]] = row["ZWITHDRAWTRANSACTION"]
        return refund_to_withdraw

    def get_tags_map(self) -> Dict[ID, List[ID]]:
        transactions_to_tags: Dict[ID, List[ID]] = defaultdict(list)
        cur = self._con.cursor()
        res = cur.execute(
            """
        SELECT Z_36TRANSACTIONS, Z_35TAGS FROM  Z_36TAGS
        
        """
        )
        for row in res.fetchall():
            transactions_to_tags[row["Z_36TRANSACTIONS"]].append(row["Z_35TAGS"])
        return transactions_to_tags

    def get_users(self) -> Dict[ID, str]:
        users_map: Dict[ID, str] = {}
        cur = self._con.cursor()
        res = cur.execute(
            """
        SELECT Z_PK, ZSYNCLOGIN FROM  "ZUSER"
        
        """
        )
        for row in res.fetchall():
            users_map[row["Z_PK"]] = row["ZSYNCLOGIN"]
        return users_map



from dataclasses import asdict
from typing import Dict
from pathlib import Path

import logging
import json
import click
import pandas as pd

from moneywiz_api import MoneywizApi
from moneywiz_api.managers.record_manager import RecordManager
from moneywiz_api.types import ID, GID


logger = logging.getLogger(__name__)


class ShellHelper:

    def __init__(self, moneywiz_api: MoneywizApi) -> None:
        self._mw_api = moneywiz_api

    def view_id(self, record_id: ID):
        record = self._mw_api.accessor.get_record(record_id)
        click.echo(self._mw_api.accessor.typename_for(record.ent()))
        click.echo(json.dumps(record.filtered(), sort_keys=True, indent=4))

    def view_gid(self, record_gid: GID):
        record = self._mw_api.accessor.get_record_by_gid(record_gid)
        click.echo(self._mw_api.accessor.typename_for(record.ent()))
        click.echo(json.dumps(record.filtered(), sort_keys=True, indent=4))

    def write_stats_data_files(self, path_prefix: Path = Path("data/stats")):
        Path(path_prefix).mkdir(parents=True, exist_ok=True)
        managers_map: Dict[str, object] = {
            "ent": self._mw_api.accessor,
            "account": self._mw_api.account_manager,
            "payee": self._mw_api.payee_manager,
            "category": self._mw_api.category_manager,
            "transaction": self._mw_api.transaction_manager,
            "investment_holding": self._mw_api.investment_holding_manager,
        }

        for name, obj in managers_map.items():
            with open(f"{path_prefix}/{name}.data", "w", encoding="utf-8") as file:
                click.echo(obj, file=file)

    def pd_table(self, manager: RecordManager) -> pd.DataFrame:
        records = manager.records().values()
        df = pd.DataFrame.from_records([r.as_dict() for r in records])
        return df

    def users_table(self) -> pd.DataFrame:
        users = self._mw_api.accessor.get_users()
        records = [
            {"id": id, "login_name": login_name} for id, login_name in users.items()
        ]
        return pd.DataFrame.from_records(records).sort_values(by=["id"], ascending=True)

    def categories_table(self, user_id: ID) -> pd.DataFrame:
        categories = self._mw_api.category_manager.get_categories_for_user(user_id)
        return pd.DataFrame.from_records(
            [category.as_dict() for category in categories]
        ).sort_values(by=["id"], ascending=True)

    def accounts_table(self, user_id: ID) -> pd.DataFrame:
        # pd.set_option("display.max_colwidth", None)
        # pd.set_option("display.max_rows", None)

        accounts = self._mw_api.account_manager.get_accounts_for_user(user_id)
        return pd.DataFrame.from_records(
            [account.as_dict() for account in accounts]
        ).sort_values(by=["group_id", "display_order"])

    def investment_holdings_table(self, account_id: ID) -> pd.DataFrame:
        investment_holdings = (
            self._mw_api.investment_holding_manager.get_holdings_for_account(account_id)
        )
        return pd.DataFrame.from_records(
            [investment_holding.as_dict() for investment_holding in investment_holdings]
        ).sort_values(by=["account", "symbol"])

    def transactions_table(self, account_id: ID) -> pd.DataFrame:
        records = self._mw_api.transaction_manager.get_all_for_account(account_id)

        return pd.DataFrame.from_records(
            [record.as_dict() for record in records]
        ).sort_values(by=["datetime"], ascending=False)



from typing import Dict, Callable

from moneywiz_api.model.payee import Payee
from moneywiz_api.managers.record_manager import RecordManager


class PayeeManager(RecordManager[Payee]):
    def __init__(self):
        super().__init__()

    @property
    def ents(self) -> Dict[str, Callable]:
        return {
            "Payee": Payee,
        }



import logging
import random
import readline
import rlcompleter
from code import InteractiveConsole
from os.path import expanduser
from pathlib import Path
from typing import Dict, List

import click
import pandas as pd

from moneywiz_api.cli.helpers import ShellHelper
from moneywiz_api.moneywiz_api import MoneywizApi


def get_default_path() -> Path:
    return Path(
        expanduser(
            "~/Library/Containers/com.moneywiz.personalfinance/Data/Documents/.AppData/ipadMoneyWiz.sqlite"
        )
    )


@click.command()
@click.argument(
    "DB_FILE_PATH",
    type=click.Path(writable=False, readable=True, exists=True),
    default=get_default_path(),
)
@click.option(
    "-d",
    "--demo-dump",
    is_flag=True,
    help="Dumy some demo data",
)
@click.option(
    "--log-level",
    default="INFO",
    type=click.Choice(
        ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"], case_sensitive=False
    ),
    help="Set the logging level (DEBUG, INFO, WARNING, ERROR, CRITICAL)",
)
def main(db_file_path, demo_dump, log_level):
    """
    Interactive shell to access MoneyWiz (Read-only)
    """

    # Configure logging level
    numeric_level = getattr(logging, log_level.upper(), None)
    if not isinstance(numeric_level, int):
        raise ValueError(f"Invalid log level: {log_level}")

    logging.basicConfig(level=numeric_level)

    root_logger = logging.getLogger()
    root_logger.setLevel(numeric_level)

    moneywiz_api = MoneywizApi(db_file_path)

    (
        accessor,
        account_manager,
        payee_manager,
        category_manager,
        transaction_manager,
        investment_holding_manager,
        tag_manager,
    ) = (
        moneywiz_api.accessor,
        moneywiz_api.account_manager,
        moneywiz_api.payee_manager,
        moneywiz_api.category_manager,
        moneywiz_api.transaction_manager,
        moneywiz_api.investment_holding_manager,
        moneywiz_api.tag_manager,
    )

    helper = ShellHelper(moneywiz_api)

    names: Dict[str, str] = {
        f"{moneywiz_api=}".split("=")[0]: "MoneyWiz API",
        f"{accessor=}".split("=")[0]: "MoneyWiz Database Accessor",
        f"{account_manager=}".split("=")[0]: "Account Manager",
        f"{payee_manager=}".split("=")[0]: "Payee Manager",
        f"{category_manager=}".split("=")[0]: "Category Manager",
        f"{transaction_manager=}".split("=")[0]: "Transaction Manager",
        f"{investment_holding_manager=}".split("=")[0]: "Investment Holding Manager",
        f"{tag_manager=}".split("=")[0]: "Tag Manager",
        f"{helper=}".split("=")[0]: "Shell Helper",
    }

    banner: List[str] = [
        f"Read-only MoneyWiz Shell on {db_file_path}",
        "",
        "Available components:",
        *[f"- {component:30}  {desc}" for component, desc in names.items()],
        "===================================================================",
    ]

    if demo_dump:
        _users_table = helper.users_table()
        click.secho("Users Table", fg="yellow")
        click.secho("--------------------------------", fg="yellow")
        click.secho(_users_table.to_string(index=False))
        click.secho("--------------------------------\n", fg="yellow")

        _userid_list = _users_table["id"].tolist()
        _userid_list.remove(1)
        _user_id = random.choice(_userid_list)

        _categories_table = helper.categories_table(_user_id)
        click.secho(f"Categories Table for User {_user_id}", fg="yellow")
        click.secho("--------------------------------", fg="yellow")
        click.secho(
            _categories_table[["id", "name", "type"]].sample(5).to_string(index=False)
        )
        click.secho("--------------------------------\n", fg="yellow")

        _accounts_table = helper.accounts_table(_user_id)
        click.secho(f"Accounts Table for User {_user_id}", fg="yellow")
        click.secho("--------------------------------", fg="yellow")
        click.secho(_accounts_table[["id", "name"]].sample(5).to_string(index=False))
        click.secho("--------------------------------\n", fg="yellow")

    _vars = globals()
    _vars.update(locals())

    pd.options.display.max_rows = None
    pd.options.display.max_colwidth = None

    readline.set_completer(rlcompleter.Completer(_vars).complete)
    readline.parse_and_bind("tab: complete")
    InteractiveConsole(_vars).interact(banner="\n".join(banner))


if __name__ == "__main__":
    main()  # pylint: disable=no-value-for-parameter



from moneywiz_api.model.record import Record

from moneywiz_api.model.account import (
    Account,
    InvestmentAccount,
    CashAccount,
    LoanAccount,
    ForexAccount,
    BankChequeAccount,
    BankSavingAccount,
    CreditCardAccount,
)

from moneywiz_api.model.category import Category
from moneywiz_api.model.investment_holding import InvestmentHolding
from moneywiz_api.model.payee import Payee
from moneywiz_api.model.tag import Tag
from moneywiz_api.model.transaction import (
    RefundTransaction,
    Transaction,
    WithdrawTransaction,
    DepositTransaction,
    TransferDepositTransaction,
    TransferBudgetTransaction,
    TransferWithdrawTransaction,
    InvestmentTransaction,
    InvestmentBuyTransaction,
    InvestmentSellTransaction,
    InvestmentExchangeTransaction,
    ReconcileTransaction,
)


__all__ = [
    "Record",
    "Account",
    "InvestmentAccount",
    "CashAccount",
    "LoanAccount",
    "ForexAccount",
    "BankChequeAccount",
    "BankSavingAccount",
    "CreditCardAccount",
    "Category",
    "InvestmentHolding",
    "Payee",
    "Tag",
    "RefundTransaction",
    "Transaction",
    "WithdrawTransaction",
    "DepositTransaction",
    "TransferDepositTransaction",
    "TransferBudgetTransaction",
    "TransferWithdrawTransaction",
    "InvestmentTransaction",
    "InvestmentBuyTransaction",
    "InvestmentSellTransaction",
    "InvestmentExchangeTransaction",
    "ReconcileTransaction",
]



from abc import ABC, abstractmethod
from typing import Dict, Generic, TypeVar, Callable

from moneywiz_api.database_accessor import DatabaseAccessor
from moneywiz_api.model.record import Record
from moneywiz_api.types import ID, GID


T = TypeVar("T", bound=Record)


class RecordManager(ABC, Generic[T]):
    def __init__(self):
        self._records: Dict[ID, T] = {}
        self._gid_to_id: Dict[GID, ID] = {}

    @property
    @abstractmethod
    def ents(self) -> Dict[str, Callable]:
        raise NotImplementedError()

    def load(self, db_accessor: DatabaseAccessor) -> None:
        records = db_accessor.query_objects(self.ents.keys())

        for record in records:
            typename = db_accessor.typename_for(record["Z_ENT"])
            if typename in self.ents:
                obj = self.ents[typename](record)
                self.add(obj)

    def add(self, record: T) -> None:
        self._records[record.id] = record
        if record.gid in self._gid_to_id:
            raise RuntimeError(
                f"Duplicate gid for {record}, existing record Id {self._gid_to_id[record.gid]}"
            )

        self._gid_to_id[record.gid] = record.id

    def get(self, record_id: ID) -> T | None:
        return self._records.get(record_id)

    def get_by_gid(self, gid: GID) -> T | None:
        return self._records.get(self._gid_to_id.get(gid))

    def records(self) -> Dict[ID, T]:
        return self._records

    def __repr__(self):
        return "\n".join(f"{key}: {value}" for key, value in self.records().items())



from datetime import datetime
from typing import Optional
from decimal import Decimal

_CUTOFF = datetime(2001, 1, 1, 0, 0, 0).timestamp()


def get_datetime(date: float) -> datetime:
    return datetime.fromtimestamp(date + _CUTOFF)


def get_date_iso(date: float) -> str:
    return get_datetime(date).date().isoformat()


def get_date(dt: datetime) -> float:
    return dt.timestamp() - _CUTOFF



from dataclasses import dataclass
from typing import Optional

from moneywiz_api.model.record import Record
from moneywiz_api.types import CategoryType, ID


@dataclass
class Category(Record):
    """
    ENT: 19
    """

    name: str
    parent_id: Optional[int]
    type: CategoryType
    user: ID

    def __init__(self, row):
        super().__init__(row)
        self.name = row["ZNAME2"]
        self.parent_id = row["ZPARENTCATEGORY"]
        self.type = self._convert_type(row["ZTYPE2"])
        self.user = row["ZUSER3"]

        # Fixes

        # Validate
        assert self.name is not None, self.as_dict()
        assert self.type is not None, self.as_dict()
        assert self.user is not None, self.as_dict()

    @staticmethod
    def _convert_type(type_: Optional[int]) -> CategoryType:
        if type_ and type_ in [1, 2]:
            return "Expenses" if type_ == 1 else "Income"
        raise RuntimeError(f"Invalid type {type_}")



from typing import Optional, Dict, Any
from datetime import datetime
from decimal import Decimal

from moneywiz_api.utils import get_datetime


class RawDataHandler:

    @staticmethod
    def get_datetime(row: Dict[str, Any], key: str) -> datetime:
        raw_value = row[key]
        assert isinstance(raw_value, float) or isinstance(
            raw_value, int
        ), f"row['{key}'] = {row[key]}, is not a float or int, where row is: " + str(
            RawDataHandler.filter_row(row)
        )
        return get_datetime(raw_value)

    @staticmethod
    def get_nullable_decimal(row: Dict[str, Any], key: str) -> Optional[Decimal]:
        raw_value = row[key]
        if raw_value is None:
            return None
        else:
            return RawDataHandler.get_decimal(row, key)

    @staticmethod
    def get_decimal(row: Dict[str, Any], key: str) -> Decimal:
        raw_value = row[key]
        assert isinstance(raw_value, float) or isinstance(
            raw_value, int
        ), f"row['{key}'] = {row[key]}, is not a float or int, where row is: " + str(
            RawDataHandler.filter_row(row)
        )
        return Decimal(str(raw_value))

    @staticmethod
    def filter_row(row: Dict[str, Any]) -> Dict[str, Any]:
        copy = {k: v for k, v in row.items()}
        del copy["ZMANUALHISTORICALPRICESPERSHARE"]
        del copy["ZIMPORTLINKIDARRAY2"]
        del copy["ZIMPORTLINKIDARRAY"]
        del copy["ZBANKLOGOPRIMARYCOLOR"]
        return {
            k: v
            for k, v in copy.items()
            if (v is not None) and (not k.startswith("Z9_"))
        }



from dataclasses import dataclass, field
from typing import Dict, Any, Optional
from decimal import Decimal

from moneywiz_api.model.raw_data_handler import RawDataHandler as RDH
from moneywiz_api.model.record import Record
from moneywiz_api.types import ID


@dataclass
class InvestmentHolding(Record):
    """
    ENT: 24
    """

    account: ID
    opening_number_of_shares: Optional[Decimal]

    number_of_shares: Decimal
    # price_per_share: Decimal
    symbol: str
    holding_type: Optional[str]
    description: str

    price_per_share_available_online: bool

    """
    Unsure about the usage.
    value can be 0,1
    
    seems like: 
        0 -> aggregate balance from all transactions
        1 -> use number_of_shares as balance
    """
    _investment_object_type: int = field(repr=False)

    """
    Unsure
    
    seems like the the cost for the shares which is not from transactions
    """
    _cost_basis_of_missing_ob_shares: Decimal = field(repr=False)

    def __init__(self, row):
        super().__init__(row)
        self.account = row["ZINVESTMENTACCOUNT"]
        self.opening_number_of_shares = RDH.get_nullable_decimal(
            row, "ZOPENNINGNUMBEROFSHARES"
        )
        self.number_of_shares = RDH.get_decimal(row, "ZNUMBEROFSHARES")
        # self.price_per_share = row["ZPRICEPERSHARE"]
        self.symbol = row["ZSYMBOL"]
        self.holding_type = row["ZHOLDINGTYPE"]
        self.description = row["ZDESC"]
        self.price_per_share_available_online = (
            row["ZISPRICEPERSHAREAVAILABLEONLINE"] == 1
        )

        self._investment_object_type = row["ZINVESTMENTOBJECTTYPE"]
        self._cost_basis_of_missing_ob_shares = RDH.get_decimal(
            row, "ZCOSTBASISOFMISSINGOBSHARES"
        )

        # Fixes

        # Validate
        self.validate()

    def validate(self):
        assert self.account is not None, self.as_dict()
        assert self.number_of_shares is not None, self.as_dict()
        # assert self.price_per_share is not None
        assert self.symbol is not None, self.as_dict()
        assert self.description is not None, self.as_dict()

        assert self._investment_object_type is not None, self.as_dict()
        assert self._cost_basis_of_missing_ob_shares is not None, self.as_dict()

    def as_dict(self) -> Dict[str, Any]:
        original = super().as_dict()
        del original["_investment_object_type"]
        del original["_cost_basis_of_missing_ob_shares"]
        return original



from abc import ABC
from dataclasses import dataclass, field
from decimal import Decimal

from moneywiz_api.types import ID
from moneywiz_api.model.raw_data_handler import RawDataHandler as RDH
from moneywiz_api.model.record import Record


@dataclass
class Account(Record, ABC):
    """
    ENT: 9
    """

    display_order: int = field(repr=False)
    group_id: int = field(repr=False)

    name: str
    currency: str
    opening_balance: Decimal  # might be a tiny number
    info: str
    user: ID

    def __init__(self, row):
        super().__init__(row)
        self.display_order = row["ZDISPLAYORDER"]
        self.group_id = row["ZGROUPID"]

        self.name = row["ZNAME"]
        self.currency = row["ZCURRENCYNAME"]
        self.opening_balance = RDH.get_decimal(row, "ZOPENINGBALANCE")
        self.info = row["ZINFO"]

        self.user = row["ZUSER"]

        # Fixes

        # Validate
        assert self.display_order is not None, self.as_dict()
        assert self.group_id is not None, self.as_dict()
        assert self.name is not None, self.as_dict()
        assert self.currency is not None, self.as_dict()
        assert self.opening_balance is not None, self.as_dict()
        assert self.info is not None, self.as_dict()
        assert self.user is not None, self.as_dict()


@dataclass
class BankChequeAccount(Account):
    """
    ENT: 10
    """

    def __init__(self, row):
        super().__init__(row)


@dataclass
class BankSavingAccount(Account):
    """
    ENT: 11
    """

    def __init__(self, row):
        super().__init__(row)


@dataclass
class CashAccount(Account):
    """
    ENT: 12
    """

    def __init__(self, row):
        super().__init__(row)


@dataclass
class CreditCardAccount(Account):
    """
    ENT: 13
    """

    statement_day: int  # day in the month

    def __init__(self, row):
        super().__init__(row)
        self.statement_day = row["ZSTATEMENTENDDAY"]

        # Validate
        assert self.statement_day is not None


@dataclass
class LoanAccount(CreditCardAccount):
    """
    ENT: 14
    """

    def __init__(self, row):
        super().__init__(row)


@dataclass
class InvestmentAccount(Account):
    """
    ENT: 15
    """

    def __init__(self, row):
        super().__init__(row)


@dataclass
class ForexAccount(InvestmentAccount):
    """
    ENT: 16
    """

    def __init__(self, row):
        super().__init__(row)



from abc import ABC
from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from typing import Optional

import pytest

from moneywiz_api.model.raw_data_handler import RawDataHandler as RDH
from moneywiz_api.model.record import Record
from moneywiz_api.types import ID


@dataclass
class Transaction(Record, ABC):
    """
    ENT: 36
    """

    reconciled: bool

    amount: Decimal
    description: str
    datetime: datetime
    notes: Optional[str]

    def __init__(self, row):
        super().__init__(row)
        self.reconciled = row["ZRECONCILED"] == 1
        self.amount = RDH.get_decimal(row, "ZAMOUNT1")
        self.description = row["ZDESC2"]
        self.datetime = RDH.get_datetime(row, "ZDATE1")
        self.notes = row["ZNOTES1"]

        # Fixes

        # Validate
        assert self.reconciled is not None, self.as_dict()
        assert self.amount is not None, self.as_dict()
        assert self.description is not None, self.as_dict()
        assert self.datetime is not None, self.as_dict()
        # self.notes can be None


@dataclass
class DepositTransaction(Transaction):
    """
    ENT: 37
    """

    account: ID
    amount: Decimal  # neg: expense, pos: income
    payee: Optional[ID]

    # FX
    original_currency: str
    original_amount: Decimal  # neg: expense, pos: income
    original_exchange_rate: Optional[Decimal]

    def __init__(self, row):
        super().__init__(row)
        self.account = row["ZACCOUNT2"]
        self.amount = RDH.get_decimal(row, "ZAMOUNT1")
        self.payee = row["ZPAYEE2"]
        self.original_currency = row["ZORIGINALCURRENCY"]
        self.original_amount = RDH.get_decimal(row, "ZORIGINALAMOUNT")
        self.original_exchange_rate = RDH.get_nullable_decimal(
            row, "ZORIGINALEXCHANGERATE"
        )

        # Fixes
        if self.original_exchange_rate == Decimal(0):
            self.original_exchange_rate = None

        # Validate
        self.validate()

    def validate(self):
        assert self.account is not None, self.as_dict()
        assert self.amount is not None, self.as_dict()
        # self.payee can be None
        assert self.original_currency is not None, self.as_dict()
        assert self.original_amount is not None, self.as_dict()

        assert self.amount * self.original_amount > 0, self.as_dict()  # Same sign
        if self.original_exchange_rate is not None:
            assert (
                self.amount == self.original_amount * self.original_exchange_rate
            ), self.as_dict()


@dataclass
class InvestmentExchangeTransaction(Transaction):
    """
    ENT: 38
    """

    def __init__(self, row):
        super().__init__(row)
        raise NotImplementedError()


@dataclass
class InvestmentTransaction(Transaction, ABC):
    """
    ENT: 39
    """

    def __init__(self, row):
        super().__init__(row)


@dataclass
class InvestmentBuyTransaction(InvestmentTransaction):
    """
    ENT: 40
    """

    account: ID
    amount: Decimal

    fee: Decimal

    investment_holding: ID
    number_of_shares: Decimal
    price_per_share: Decimal

    def __init__(self, row):
        super().__init__(row)
        self.account = row["ZACCOUNT2"]
        self.amount = RDH.get_decimal(row, "ZAMOUNT1")

        self.fee = RDH.get_decimal(row, "ZFEE2")

        self.investment_holding = row["ZINVESTMENTHOLDING"]
        self.number_of_shares = RDH.get_decimal(row, "ZNUMBEROFSHARES1")
        self.price_per_share = RDH.get_decimal(row, "ZPRICEPERSHARE1")

        # Fixes
        self.fee = max(self.fee, 0)

        # Validate
        self.validate()

    def validate(self):
        assert self.account is not None
        assert self.amount is not None
        assert self.amount <= 0
        assert self.fee is not None
        assert self.fee >= 0
        # Either tiny (close to 0) or positive
        assert abs(self.fee) == pytest.approx(0, abs=0.001) or self.fee > 0.001
        assert self.investment_holding is not None
        assert self.number_of_shares is not None
        assert self.number_of_shares > 0
        assert self.price_per_share is not None
        assert self.price_per_share >= 0
        assert -(
            self.number_of_shares * self.price_per_share + self.fee
        ) == pytest.approx(self.amount, abs=0.001)


@dataclass
class InvestmentSellTransaction(InvestmentTransaction):
    """
    ENT: 41
    """

    account: ID
    amount: Decimal  # neg: loss after fees, pos: income

    fee: Decimal

    investment_holding: ID
    number_of_shares: Decimal
    price_per_share: Decimal

    def __init__(self, row):
        super().__init__(row)
        self.account = row["ZACCOUNT2"]
        self.amount = RDH.get_decimal(row, "ZAMOUNT1")

        self.fee = RDH.get_decimal(row, "ZFEE2")

        self.investment_holding = row["ZINVESTMENTHOLDING"]
        self.number_of_shares = RDH.get_decimal(row, "ZNUMBEROFSHARES1")
        self.price_per_share = RDH.get_decimal(row, "ZPRICEPERSHARE1")

        # Fixes
        self.fee = max(self.fee, 0)

        # Validate
        self.validate()

    def validate(self):
        assert self.account is not None
        assert self.amount is not None

        assert self.fee is not None
        assert self.fee >= 0
        # Either tiny (close to 0) or positive
        assert abs(self.fee) == pytest.approx(0, abs=0.001) or self.fee > 0.001

        assert self.investment_holding is not None
        assert self.number_of_shares is not None
        assert self.number_of_shares > 0
        assert self.price_per_share is not None
        assert self.price_per_share >= 0
        assert (
            self.number_of_shares * self.price_per_share - self.fee
        ) == pytest.approx(self.amount, abs=0.001)


@dataclass
class ReconcileTransaction(Transaction):
    """
    ENT: 42
    """

    account: ID
    amount: Decimal  # neg: expense, pos: income
    reconcile_amount: Decimal  # new balance

    def __init__(self, row):
        super().__init__(row)
        self.account = row["ZACCOUNT2"]
        self.amount = RDH.get_decimal(row, "ZAMOUNT1")
        self.reconcile_amount = RDH.get_decimal(row, "ZRECONCILEAMOUNT")

        # Validate
        self.validate()

    def validate(self):
        assert self.account is not None
        assert self.amount is not None
        assert self.reconcile_amount is not None


@dataclass
class RefundTransaction(Transaction):
    """
    ENT: 43
    """

    account: ID
    amount: Decimal
    payee: Optional[ID]

    # FX
    original_currency: str
    original_amount: Decimal
    original_exchange_rate: Optional[Decimal]

    def __init__(self, row):
        super().__init__(row)
        self.account = row["ZACCOUNT2"]
        self.amount = RDH.get_decimal(row, "ZAMOUNT1")
        self.payee = row["ZPAYEE2"]

        self.original_currency = row["ZORIGINALCURRENCY"]
        self.original_amount = RDH.get_decimal(row, "ZORIGINALAMOUNT")
        self.original_exchange_rate = RDH.get_nullable_decimal(
            row, "ZORIGINALEXCHANGERATE"
        )

        # Fixes
        if self.original_exchange_rate == Decimal(0):
            self.original_exchange_rate = None

        # Validate
        self.validate()

    def validate(self):
        assert self.account is not None
        assert self.amount is not None
        assert self.amount > 0

        assert self.original_currency is not None
        assert self.original_amount is not None
        assert self.original_amount > 0

        if self.original_exchange_rate is not None:
            assert self.amount == pytest.approx(
                self.original_amount * self.original_exchange_rate, abs=0.001
            )


@dataclass
class TransferBudgetTransaction(Transaction):
    """
    ENT: 44
    """

    def __init__(self, row):
        super().__init__(row)
        # TODO: Not Implemented


@dataclass
class TransferDepositTransaction(Transaction):
    """
    ENT: 45
    """

    account: ID
    amount: Decimal  # pos: in

    sender_account: ID
    sender_transaction: ID

    original_amount: Decimal  # ATTENTION: sign got fixed
    original_currency: str

    sender_amount: Decimal
    sender_currency: str

    original_fee: Optional[Decimal]
    original_fee_currency: Optional[str]

    original_exchange_rate: Decimal

    def __init__(self, row):
        super().__init__(row)
        self.account = row["ZACCOUNT2"]
        self.amount = RDH.get_decimal(row, "ZAMOUNT1")

        self.sender_account = row["ZSENDERACCOUNT"]
        self.sender_transaction = row["ZSENDERTRANSACTION"]

        self.original_amount = RDH.get_decimal(row, "ZORIGINALAMOUNT")
        self.original_currency = row["ZORIGINALCURRENCY"]
        self.sender_amount = RDH.get_decimal(row, "ZORIGINALSENDERAMOUNT")
        self.sender_currency = row["ZORIGINALSENDERCURRENCY"]

        self.original_fee = RDH.get_nullable_decimal(row, "ZORIGINALFEE")
        self.original_fee_currency = row["ZORIGINALFEECURRENCY"]

        self.original_exchange_rate = RDH.get_decimal(row, "ZORIGINALEXCHANGERATE")

        # Fixes
        self.original_amount = abs(self.original_amount)

        # Validate
        self.validate()

    def validate(self):
        assert self.account is not None
        assert self.amount is not None
        assert self.amount > 0
        assert self.sender_account is not None
        assert self.sender_transaction is not None
        assert self.original_amount is not None
        assert self.original_amount > 0
        assert self.original_currency is not None
        assert self.sender_amount is not None
        assert self.sender_amount <= 0
        assert self.sender_currency is not None

        if self.original_fee is not None and self.original_fee != 0:
            assert self.original_fee_currency is not None

        assert self.original_exchange_rate is not None

        # assert self.amount ==  self.original_amount # original_amount could be different with amount ZCURRENCYEXCHANGERATE is playing up
        assert self.original_amount == pytest.approx(
            -self.sender_amount * self.original_exchange_rate
            - (self.original_fee or 0),
            abs=0.001,
        )


@dataclass
class TransferWithdrawTransaction(Transaction):
    """
    ENT: 46
    """

    account: ID
    amount: Decimal  # neg: out

    recipient_account: ID
    recipient_transaction: ID

    original_amount: Decimal  # always neg
    original_currency: str

    recipient_amount: Decimal  # ATTENTION: sign got fixed
    recipient_currency: str

    original_fee: Optional[Decimal]
    original_fee_currency: Optional[str]

    original_exchange_rate: Decimal

    def __init__(self, row):
        super().__init__(row)
        self.account = row["ZACCOUNT2"]
        self.amount = RDH.get_decimal(row, "ZAMOUNT1")

        self.recipient_account = row["ZRECIPIENTACCOUNT1"]
        self.recipient_transaction = row["ZRECIPIENTTRANSACTION"]

        self.original_amount = RDH.get_decimal(row, "ZORIGINALAMOUNT")
        self.original_currency = row["ZORIGINALCURRENCY"]
        self.recipient_amount = RDH.get_decimal(row, "ZORIGINALRECIPIENTAMOUNT")
        self.recipient_currency = row["ZORIGINALRECIPIENTCURRENCY"]

        self.original_fee = RDH.get_nullable_decimal(row, "ZORIGINALFEE")
        self.original_fee_currency = row["ZORIGINALFEECURRENCY"]

        self.original_exchange_rate = RDH.get_decimal(row, "ZORIGINALEXCHANGERATE")

        # Fixes
        self.recipient_amount = abs(self.recipient_amount)

        # Validate
        self.validate()

    def validate(self):
        assert self.account is not None
        assert self.amount is not None
        assert self.amount < 0
        assert self.recipient_account is not None
        assert self.recipient_transaction is not None
        assert self.original_amount is not None
        assert self.original_amount < 0
        assert self.original_currency is not None
        assert self.recipient_amount is not None
        assert self.recipient_amount > 0
        assert self.recipient_currency is not None

        if self.original_fee is not None and self.original_fee != 0:
            assert self.original_fee_currency is not None

        assert self.original_exchange_rate is not None

        assert self.amount == self.original_amount
        assert self.amount == pytest.approx(
            -self.recipient_amount / self.original_exchange_rate,
            abs=0.001,
        )


@dataclass
class WithdrawTransaction(Transaction):
    """
    ENT: 47
    """

    account: ID
    amount: Decimal  # neg: expense, pos: income
    payee: Optional[ID]

    # FX
    original_currency: str
    original_amount: Decimal  # neg: expense, pos: income ATTENTION: sign got fixed
    original_exchange_rate: Optional[Decimal]

    def __init__(self, row):
        super().__init__(row)
        self.account = row["ZACCOUNT2"]
        self.amount = RDH.get_decimal(row, "ZAMOUNT1")
        self.payee = row["ZPAYEE2"]

        self.original_currency = row["ZORIGINALCURRENCY"]
        self.original_amount = RDH.get_decimal(row, "ZORIGINALAMOUNT")
        self.original_exchange_rate = RDH.get_nullable_decimal(
            row, "ZORIGINALEXCHANGERATE"
        )

        # Fixes
        if self.amount * self.original_amount < 0:
            self.original_amount = -self.original_amount

        if self.original_exchange_rate == Decimal(0):
            self.original_exchange_rate = None

        # Validate
        self.validate()

    def validate(self):
        assert self.account is not None
        assert self.amount is not None
        # self.payee can be None
        assert self.original_currency is not None
        assert self.original_amount is not None

        assert self.amount * self.original_amount > 0

        if self.original_exchange_rate is not None:
            assert self.amount == pytest.approx(
                self.original_amount * self.original_exchange_rate, abs=0.001
            )



from pathlib import Path
import logging

from moneywiz_api.database_accessor import DatabaseAccessor
from moneywiz_api.managers.account_manager import AccountManager
from moneywiz_api.managers.category_manager import CategoryManager
from moneywiz_api.managers.investment_holding_manager import (
    InvestmentHoldingManager,
)
from moneywiz_api.managers.payee_manager import PayeeManager
from moneywiz_api.managers.transaction_manager import TransactionManager
from moneywiz_api.managers.tag_manager import TagManager

logger = logging.getLogger(__name__)


class MoneywizApi:
    def __init__(self, db_file: Path):
        self.accessor = DatabaseAccessor(db_file)
        self.account_manager = AccountManager()
        self.payee_manager = PayeeManager()
        self.category_manager = CategoryManager()
        self.transaction_manager = TransactionManager()
        self.investment_holding_manager = InvestmentHoldingManager()
        self.tag_manager = TagManager()

        self.load()

    def load(self):
        self.account_manager.load(self.accessor)
        self.payee_manager.load(self.accessor)
        self.category_manager.load(self.accessor)
        self.transaction_manager.load(self.accessor)
        self.investment_holding_manager.load(self.accessor)
        self.tag_manager.load(self.accessor)
